#!/usr/bin/python2.7
from __future__ import absolute_import, unicode_literals

import datetime
import imp
import logging
import os
import re
import socket
import traceback
import warnings
from builtins import object, str
from collections import defaultdict

import boto3
import numpy as np

from atm.classifier import Model
from atm.constants import CUSTOM_CLASS_REGEX, SELECTORS, TUNERS
from atm.database import ClassifierStatus, db_session
from atm.utilities import ensure_directory, get_instance, save_metrics, save_model, update_params

# shhh
warnings.filterwarnings('ignore')

# for garrays
os.environ['GNUMPY_IMPLICIT_CONVERSION'] = 'allow'

# load the library-wide logger
LOGGER = logging.getLogger('atm')

# Local hostname, for logging.
HOSTNAME = socket.gethostname()


# Exception thrown when something goes wrong for the worker, but the worker
# handles the error.
class ClassifierError(Exception):
    pass


class Worker(object):
    def __init__(self, database, datarun, save_files=True, cloud_mode=False,
                 aws_access_key=None, aws_secret_key=None, s3_bucket=None, s3_folder=None,
                 models_dir='models', metrics_dir='metrics', verbose_metrics=False):

        self.db = database
        self.datarun = datarun
        self.save_files = save_files
        self.cloud_mode = cloud_mode

        self.aws_access_key = aws_access_key
        self.aws_secret_key = aws_secret_key
        self.s3_bucket = s3_bucket
        self.s3_folder = s3_folder

        self.models_dir = models_dir
        self.metrics_dir = metrics_dir
        self.verbose_metrics = verbose_metrics
        ensure_directory(self.models_dir)
        ensure_directory(self.metrics_dir)

        # load the Dataset from the database
        self.dataset = self.db.get_dataset(self.datarun.dataset_id)

        # load the Selector and Tuner classes specified by our datarun
        self.load_selector()
        self.load_tuner()

    def load_selector(self):
        """
        Load and initialize the BTB class which will be responsible for
        selecting hyperpartitions.
        """
        # selector will either be a key into SELECTORS or a path to
        # a file that defines a class called CustomSelector.
        if self.datarun.selector in SELECTORS:
            Selector = SELECTORS[self.datarun.selector]
        else:
            path, classname = re.match(CUSTOM_CLASS_REGEX,
                                       self.datarun.selector).groups()
            mod = imp.load_source('btb.selection.custom', path)
            Selector = getattr(mod, classname)

        LOGGER.info('Selector: %s' % Selector)

        # generate the arguments we need to initialize the selector
        hyperpartitions = self.db.get_hyperpartitions(datarun_id=self.datarun.id)
        hp_by_method = defaultdict(list)
        for hp in hyperpartitions:
            hp_by_method[hp.method].append(hp.id)

        hyperpartition_ids = [hp.id for hp in hyperpartitions]

        # Selector classes support passing in redundant arguments
        self.selector = get_instance(Selector,
                                     choices=hyperpartition_ids,
                                     k=self.datarun.k_window,
                                     by_algorithm=dict(hp_by_method))

    def load_tuner(self):
        """
        Load, but don't initialize, the BTB class which will be responsible for
        choosing non-hyperpartition hyperparameter values (a subclass of Tuner). The
        tuner must be initialized with information about the hyperpartition, so it
        cannot be created until later.
        """
        # tuner will either be a key into TUNERS or a path to
        # a file that defines a class called CustomTuner.
        if self.datarun.tuner in TUNERS:
            self.Tuner = TUNERS[self.datarun.tuner]
        else:
            path, classname = re.match(CUSTOM_CLASS_REGEX, self.datarun.tuner).groups()
            mod = imp.load_source('btb.tuning.custom', path)
            self.Tuner = getattr(mod, classname)

        LOGGER.info('Tuner: %s' % self.Tuner)

    def select_hyperpartition(self):
        """
        Use the hyperpartition selection method specified by our datarun to choose a
        hyperpartition of hyperparameters from the ModelHub. Only consider
        partitions for which gridding is not complete.
        """
        hyperpartitions = self.db.get_hyperpartitions(datarun_id=self.datarun.id)

        # load classifiers and build scores lists
        # make sure all hyperpartitions are present in the dict, even ones that
        # don't have any classifiers. That way the selector can choose hyperpartitions
        # that haven't been scored yet.
        hyperpartition_scores = {fs.id: [] for fs in hyperpartitions}
        classifiers = self.db.get_classifiers(datarun_id=self.datarun.id)

        for c in classifiers:
            # ignore hyperpartitions for which gridding is done
            if c.hyperpartition_id not in hyperpartition_scores:
                continue

            # the cast to float is necessary because the score is a Decimal;
            # doing Decimal-float arithmetic throws errors later on.
            score = float(getattr(c, self.datarun.score_target) or 0)
            hyperpartition_scores[c.hyperpartition_id].append(score)

        hyperpartition_id = self.selector.select(hyperpartition_scores)
        return self.db.get_hyperpartition(hyperpartition_id)

    def tune_hyperparameters(self, hyperpartition):
        """
        Use the hyperparameter tuning method specified by our datarun to choose
        a set of hyperparameters from the potential space.
        """
        # Get parameter metadata for this hyperpartition
        tunables = hyperpartition.tunables

        # If there aren't any tunable parameters, we're done. Return the vector
        # of values in the hyperpartition and mark the set as finished.
        if not len(tunables):
            LOGGER.warning('No tunables for hyperpartition %d' % hyperpartition.id)
            self.db.mark_hyperpartition_gridding_done(hyperpartition.id)
            return update_params(params=[],
                                 tunables=tunables,
                                 categoricals=hyperpartition.categoricals,
                                 constants=hyperpartition.constants)

        # Get previously-used parameters: every classifier should either be
        # completed or have thrown an error
        all_clfs = self.db.get_classifiers(hyperpartition_id=hyperpartition.id)
        classifiers = [c for c in all_clfs if c.status == ClassifierStatus.COMPLETE]

        X = [c.hyperparameter_values for c in classifiers]
        y = np.array([float(getattr(c, self.datarun.score_target)) for c in classifiers])

        # Initialize the tuner and propose a new set of parameters
        # this has to be initialized with information from the hyperpartition, so we
        # need to do it fresh for each classifier (not in load_tuner)
        tuner = get_instance(self.Tuner,
                             tunables=tunables,
                             gridding=self.datarun.gridding,
                             r_minimum=self.datarun.r_minimum)
        if len(X) > 0:
            tuner.add(X, y)

        params = tuner.propose()
        if params is None and self.datarun.gridding:
            LOGGER.info('Gridding done for hyperpartition %d' % hyperpartition.id)
            self.db.mark_hyperpartition_gridding_done(hyperpartition.id)
            return None

        # Append categorical and constants to the params.
        return update_params(params=params,
                             categoricals=hyperpartition.categoricals,
                             constants=hyperpartition.constants)

    def test_classifier(self, method, params):
        """
        Given a set of fully-qualified hyperparameters, create and test a
        classifier model.
        Returns: Model object and metrics dictionary
        """
        model = Model(method=method, params=params,
                      judgment_metric=self.datarun.metric,
                      class_column=self.dataset.class_column,
                      verbose_metrics=self.verbose_metrics)

        train_path, test_path = self.dataset.load()

        metrics = model.train_test(self.dataset)

        target = self.datarun.score_target

        def metric_string(model):
            if 'cv' in target or 'mu_sigma' in target:
                return '%.3f +- %.3f' % (model.cv_judgment_metric,
                                         2 * model.cv_judgment_metric_stdev)
            else:
                return '%.3f' % model.test_judgment_metric

        LOGGER.info('Judgment metric (%s, %s): %s' % (self.datarun.metric,
                                                      target[:-len('_judgment_metric')],
                                                      metric_string(model)))

        old_best = self.db.get_best_classifier(datarun_id=self.datarun.id,
                                               score_target=target)
        if old_best is not None:
            if getattr(model, target) > getattr(old_best, target):
                LOGGER.info('New best score! Previous best (classifier %s): %s',
                            old_best.id, metric_string(old_best))
            else:
                LOGGER.info('Best so far (classifier %s): %s',
                            old_best.id, metric_string(old_best))

        return model, metrics

    def save_classifier(self, classifier_id, model, metrics):
        """
        Update a classifier with metrics and model information and mark it as
        "complete"

        classifier_id: ID of the classifier to save
        model: Model object containing a serializable representation of the
            final model generated by this classifier.
        metrics: Dictionary containing cross-validation and test metrics data
            for the model.
        """
        # whether to save model and metrics data to the filesystem
        if self.save_files:
            # keep a database session open so that the utility functions can
            # access the linked hyperpartitions and dataruns
            with db_session(self.db):
                classifier = self.db.get_classifier(classifier_id)
                model_path = save_model(classifier, self.models_dir, model)
                metric_path = save_metrics(classifier, self.metrics_dir, metrics)

            # if necessary, save model and metrics to Amazon S3 bucket
            if self.cloud_mode:
                try:
                    model_path, metric_path = self.save_classifier_cloud(model_path, metric_path)

                except Exception:
                    msg = traceback.format_exc()
                    LOGGER.error('Error in save_classifier_cloud()')
                    self.db.mark_classifier_errored(classifier_id, error_message=msg)
        else:
            model_path = None
            metric_path = None

        # update the classifier in the database
        self.db.complete_classifier(classifier_id=classifier_id,
                                    model_location=model_path,
                                    metrics_location=metric_path,
                                    cv_score=model.cv_judgment_metric,
                                    cv_stdev=model.cv_judgment_metric_stdev,
                                    test_score=model.test_judgment_metric)

        # update this session's hyperpartition entry
        LOGGER.info('Saved classifier %d.' % classifier_id)

    def save_classifier_cloud(self, local_model_path, local_metric_path, delete_local=False):
        """
        Save a classifier to the S3 bucket supplied on __init__. Saves a
        serialized representaion of the model as well as a detailed set
        of metrics.

        local_model_path: path to serialized model in the local file system
        local_metric_path: path to serialized metrics in the local file system
        """

        client = boto3.client(
            's3',
            aws_access_key_id=self.aws_access_key,
            aws_secret_access_key=self.aws_secret_key,
        )

        if self.s3_folder:
            aws_model_path = os.path.join(self.s3_folder, local_model_path)
            aws_metric_path = os.path.join(self.s3_folder, local_metric_path)
        else:
            aws_model_path = local_model_path
            aws_metric_path = local_metric_path

        LOGGER.info('Uploading model at %s to s3://%s/%s',
                    local_model_path, self.s3_bucket, aws_model_path)
        client.upload_file(local_model_path, self.s3_bucket, aws_model_path)

        LOGGER.info('Uploading metric at %s to s3://%s/%s',
                    local_metric_path, self.s3_bucket, aws_metric_path)
        client.upload_file(local_metric_path, self.s3_bucket, aws_metric_path)

        if delete_local:
            LOGGER.info('Deleting local copies of %s and %s',
                        local_model_path, local_metric_path)
            os.remove(local_model_path)
            os.remove(local_metric_path)

        return (
            's3://{}/{}'.format(self.s3_bucket, aws_model_path),
            's3://{}/{}'.format(self.s3_bucket, aws_metric_path)
        )

    def is_datarun_finished(self):
        """
        Check to see whether the datarun is finished. This could be due to the
        budget being exhausted or due to hyperparameter gridding being done.
        """
        hyperpartitions = self.db.get_hyperpartitions(datarun_id=self.datarun.id)
        if not hyperpartitions:
            LOGGER.warning('No incomplete hyperpartitions for datarun %d present in database.'
                           % self.datarun.id)
            return True

        if self.datarun.budget_type == 'classifier':
            # hyperpartition classifier counts are updated whenever a classifier
            # is created, so this will count running, errored, and complete.
            n_completed = len(self.db.get_classifiers(datarun_id=self.datarun.id))
            if n_completed >= self.datarun.budget:
                LOGGER.warning('Classifier budget has run out!')
                return True

        elif self.datarun.budget_type == 'walltime':
            deadline = self.datarun.deadline
            if datetime.datetime.now() > deadline:
                LOGGER.warning('Walltime budget has run out!')
                return True

        return False

    def run_classifier(self, hyperpartition_id=None):
        """
        Choose hyperparameters, then use them to test and save a Classifier.
        """
        # check to see if our work is done
        if self.is_datarun_finished():
            # marked the run as done successfully
            self.db.mark_datarun_complete(self.datarun.id)
            LOGGER.warning('Datarun %d has ended.' % self.datarun.id)
            return

        try:
            LOGGER.debug('Choosing hyperparameters...')
            if hyperpartition_id is not None:
                hyperpartition = self.db.get_hyperpartition(hyperpartition_id)
                if hyperpartition.datarun_id != self.datarun.id:
                    LOGGER.error('Hyperpartition %d is not a part of datarun %d'
                                 % (hyperpartition_id, self.datarun.id))
                    return
            else:
                # use the multi-arm bandit to choose which hyperpartition to use next
                hyperpartition = self.select_hyperpartition()

            # use tuner to choose a set of parameters for the hyperpartition
            params = self.tune_hyperparameters(hyperpartition)

        except Exception:
            LOGGER.error('Error choosing hyperparameters: datarun=%s' % str(self.datarun))
            LOGGER.error(traceback.format_exc())
            raise ClassifierError()

        if params is None:
            LOGGER.warning('No parameters chosen: hyperpartition %d is finished.'
                           % hyperpartition.id)
            return

        param_info = 'Chose parameters for method "%s":' % hyperpartition.method
        for k in sorted(params.keys()):
            param_info += '\n\t%s = %s' % (k, params[k])

        LOGGER.info(param_info)

        LOGGER.debug('Creating classifier...')
        classifier = self.db.start_classifier(hyperpartition_id=hyperpartition.id,
                                              datarun_id=self.datarun.id,
                                              host=HOSTNAME,
                                              hyperparameter_values=params)

        try:
            LOGGER.debug('Testing classifier...')
            model, metrics = self.test_classifier(hyperpartition.method, params)
            LOGGER.debug('Saving classifier...')
            self.save_classifier(classifier.id, model, metrics)

        except Exception:
            msg = traceback.format_exc()
            LOGGER.error('Error testing classifier: datarun=%s' % str(self.datarun))
            LOGGER.error(msg)
            self.db.mark_classifier_errored(classifier.id, error_message=msg)
            raise ClassifierError()



import logging
import os
import traceback

import flask

LOGGER = logging.getLogger(__name__)


def make_absolute(url):
    if str(url).startswith('sqlite:///'):
        url = 'sqlite:///' + os.path.abspath(url.database)

    return url


def abort(code, message=None, error=None):
    if error is not None:
        error = traceback.format_exception_only(type(error), error)[0].strip()

    response = flask.jsonify({
        'status': code,
        'error': error,
        'message': message
    })
    response.status_code = code
    flask.abort(response)


def auto_abort(exceptions):
    def outer(function):
        def inner(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except exceptions as ex:
                abort(400, error=ex)
            except Exception as ex:
                LOGGER.exception('Uncontrolled Exception Caught')
                abort(500, error=ex)

        return inner

    return outer



from __future__ import absolute_import, unicode_literals

import json
import os
from builtins import object, range
from builtins import str as newstr

import btb

from atm.constants import METHODS


class HyperParameter(object):
    @property
    def is_categorical(self):
        return False

    @property
    def is_constant(self):
        return False


class Numeric(HyperParameter):
    def __init__(self, name, type, range):
        self.name = name
        self.type = type
        self.range = range

    @property
    def is_constant(self):
        return len(self.range) == 1

    def as_tunable(self):
        return btb.HyperParameter(param_type=self.type, param_range=self.range)


class Categorical(HyperParameter):
    def __init__(self, name, type, values):
        self.name = name
        self.type = type
        for i, val in enumerate(values):
            if val is None:
                # the value None is allowed for every parameter type
                continue
            if self.type == 'int_cat':
                values[i] = int(val)
            elif self.type == 'float_cat':
                values[i] = float(val)
            elif self.type == 'string':
                # this is necessary to avoid a bug in sklearn, which won't be
                # fixed until 0.20
                values[i] = str(newstr(val))
            elif self.type == 'bool':
                values[i] = bool(val)
        self.values = values

    @property
    def is_categorical(self):
        return True

    @property
    def is_constant(self):
        return len(self.values) == 1

    def as_tunable(self):
        return btb.HyperParameter(param_type=self.type, param_range=self.values)


class List(HyperParameter):
    def __init__(self, name, type, list_length, element):
        self.name = name
        self.length = Categorical('len(%s)' % self.name, 'int_cat', list_length)
        element_type = HYPERPARAMETER_TYPES[element['type']]
        self.element = element_type('element', **element)

    @property
    def is_categorical(self):
        return True

    def get_elements(self):
        elements = []
        for i in range(max(self.length.values)):
            # generate names for the pseudo-hyperparameters in the list
            elt_name = '%s[%d]' % (self.name, i)
            elements.append(elt_name)

        conditions = {str(i): elements[:i] for i in self.length.values}
        return elements, conditions


class HyperPartition(object):
    """
    Class which holds the hyperparameter settings that define a hyperpartition.
    """

    def __init__(self, categoricals, constants, tunables):
        """
        categoricals: the values for this hyperpartition which have been fixed
            and define the hyperpartition. List of tuples of the form ('param', val).
        constants: the values for this hyperpartition for which there was no
            choice. List of tuples of the form ('param', val).
        tunables: the free variables which must be tuned. List of tuples of the
            form ('param', HyperParameter).
        """
        self.categoricals = categoricals
        self.constants = constants
        self.tunables = tunables

    def __repr__(self):
        cats, cons, tuns = [None] * 3
        if self.categoricals:
            cats = '[%s]' % ', '.join(['%s=%s' % c for c in self.categoricals])
        if self.constants:
            cons = '[%s]' % ', '.join(['%s=%s' % c for c in self.constants])
        if self.tunables:
            tuns = '[%s]' % ', '.join(['%s' % t for t, _ in self.tunables])
        return ('<HyperPartition: categoricals: %s; constants: %s; tunables: %s>'
                % (cats, cons, tuns))


HYPERPARAMETER_TYPES = {
    'int': Numeric,
    'int_exp': Numeric,
    'float': Numeric,
    'float_exp': Numeric,
    'int_cat': Categorical,
    'float_cat': Categorical,
    'string': Categorical,
    'bool': Categorical,
    'list': List,
}


class Method(object):
    """
    This class is initialized with the name of a json configuration file.
    The config contains information about a classification method and the
    hyperparameter arguments it needs to run. Its main purpose is to generate
    hyperpartitions (possible combinations of categorical hyperparameters).
    """

    def __init__(self, method):
        """
        method: method code or path to JSON file containing all the information
            needed to specify this enumerator.
        """
        if method in METHODS:
            # if the configured method is a code, look up the path to its json
            config_path = os.path.join(os.path.dirname(__file__), 'methods', METHODS[method])
        else:
            # otherwise, it must be a path to a file
            config_path = method

        with open(config_path) as f:
            config = json.load(f)

        self.name = config['name']
        self.root_params = config['root_hyperparameters']
        self.conditions = config['conditional_hyperparameters']
        self.class_path = config['class']

        # create hyperparameters from the parameter config
        self.parameters = {}
        for k, v in list(config['hyperparameters'].items()):
            param_type = HYPERPARAMETER_TYPES[v['type']]
            self.parameters[k] = param_type(name=k, **v)

        # List hyperparameters are special. These are replaced in the
        # CPT with a size hyperparameter and sets of element hyperparameters
        # conditioned on the size.
        for name, param in list(self.parameters.items()):
            if isinstance(param, List):
                elements, conditions = param.get_elements()
                for e in elements:
                    self.parameters[e] = param.element

                # add the size parameter, remove the list parameter
                self.parameters[param.length.name] = param.length
                del self.parameters[param.name]

                # if this is a root param, replace its name with the new size
                # name in the root params list
                if param.name in self.root_params:
                    self.root_params.append(param.length.name)
                    self.root_params.remove(param.name)

                # if this is a conditional param, replace it there instead
                for var, cond in list(self.conditions.items()):
                    for val, deps in list(cond.items()):
                        if param.name in deps:
                            deps.append(param.length.name)
                            deps.remove(param.name)
                            self.conditions[var][val] = deps

                # finally, add all the potential sets of list elements as
                # conditions of the list's size
                self.conditions[param.length.name] = conditions

    def _sort_parameters(self, params):
        """
        Sort a list of HyperParameter objects into lists of constants,
        categoricals, and tunables.
        """
        constants = []
        categoricals = []
        tunables = []
        for p in params:
            param = self.parameters[p]
            if param.is_constant:
                if param.is_categorical:
                    constants.append((p, param.values[0]))
                else:
                    constants.append((p, param.range[0]))
            elif param.is_categorical:
                categoricals.append(p)
            else:
                tunables.append((p, param.as_tunable()))

        return constants, categoricals, tunables

    def _enumerate(self, fixed_cats, constants, free_cats, tunables):
        """
        Some things are fixed. Make a choice from the things that aren't fixed
        and see where that leaves us. Recurse.

        fixed_cats: a list of (name, value) tuples of qualified categorical
            variables
        constants: a list of (name, value) tuples of fixed constants
        free_cats: a list of names of free categorical variables
        tunables: a list of names of free tunable parameters

        Returns: a list of HyperPartition objects
        """
        # if there are no more free variables, we have a new HyperPartition. We've
        # reached the bottom of the recursion, so return.
        if not free_cats:
            return [HyperPartition(fixed_cats, constants, tunables)]

        parts = []

        # fix a single categorical parameter, removing it from the list of free
        # variables, and see where that takes us
        cat = free_cats.pop(0)

        for val in self.parameters[cat].values:
            # add this value to the list of qualified categoricals
            new_fixed_cats = fixed_cats + [(cat, val)]

            # these lists are copied for now
            new_constants = constants[:]
            new_free_cats = free_cats[:]
            new_tunables = tunables[:]

            # check if choosing this value opens up new parts of the conditional
            # parameter tree.
            # we need to check conditions for str(val) because all keys in json
            # must be strings.
            if cat in self.conditions and str(val) in self.conditions[cat]:
                # categorize the conditional variables which are now in play
                new_params = self.conditions[cat][str(val)]
                cons, cats, tuns = self._sort_parameters(new_params)
                new_constants = constants + cons
                new_free_cats = free_cats + cats
                new_tunables = tunables + tuns

            # recurse with the newly qualified categorical as a constant
            parts.extend(self._enumerate(fixed_cats=new_fixed_cats,
                                         constants=new_constants,
                                         free_cats=new_free_cats,
                                         tunables=new_tunables))

        return parts

    def get_hyperpartitions(self):
        """
        Traverse the CPT and enumerate all possible hyperpartitions of
        categorical parameters for this method.
        """
        return self._enumerate([], *self._sort_parameters(self.root_params))



# -*- coding: utf-8 -*-

"""Core ATM module.

This module contains the ATM class, which is the one responsible for
executing and orchestrating the main ATM functionalities.
"""

import logging
import random
import time
from datetime import datetime, timedelta
from operator import attrgetter

from tqdm import tqdm

from atm.constants import TIME_FMT, PartitionStatus, RunStatus
from atm.database import Database
from atm.method import Method
from atm.worker import ClassifierError, Worker

LOGGER = logging.getLogger(__name__)


class ATM(object):

    LOOP_WAIT = 5

    def __init__(
        self,

        # SQL Conf
        dialect='sqlite',
        database='atm.db',
        username=None,
        password=None,
        host=None,
        port=None,
        query=None,

        # AWS Conf
        access_key=None,
        secret_key=None,
        s3_bucket=None,
        s3_folder=None,

        # Log Conf
        models_dir='models',
        metrics_dir='metrics',
        verbose_metrics=False,
    ):

        self.db = Database(dialect, database, username, host, port, query)
        self.aws_access_key = access_key
        self.aws_secret_key = secret_key
        self.s3_bucket = s3_bucket
        self.s3_folder = s3_folder

        self.models_dir = models_dir
        self.metrics_dir = metrics_dir
        self.verbose_metrics = verbose_metrics

    def add_dataset(self, train_path, test_path=None, name=None,
                    description=None, class_column=None):
        return self.db.create_dataset(
            train_path=train_path,
            test_path=test_path,
            name=name,
            description=description,
            class_column=class_column,
            aws_access_key=self.aws_access_key,
            aws_secret_key=self.aws_secret_key,
        )

    def add_datarun(self, dataset_id, budget=100, budget_type='classifier',
                    gridding=0, k_window=3, metric='f1', methods=['logreg', 'dt', 'knn'],
                    r_minimum=2, run_per_partition=False, score_target='cv', priority=1,
                    selector='uniform', tuner='uniform', deadline=None):

        dataruns = list()

        if deadline:
            deadline = datetime.strptime(deadline, TIME_FMT)
            budget_type = 'walltime'

        elif budget_type == 'walltime':
            deadline = datetime.now() + timedelta(minutes=budget)

        run_description = '___'.join([tuner, selector])
        target = score_target + '_judgment_metric'

        method_parts = {}
        for method in methods:
            # enumerate all combinations of categorical variables for this method
            method_instance = Method(method)
            method_parts[method] = method_instance.get_hyperpartitions()

            LOGGER.info('method {} has {} hyperpartitions'.format(
                method, len(method_parts[method])))

        if not run_per_partition:
            datarun = self.db.create_datarun(
                dataset_id=dataset_id,
                description=run_description,
                tuner=tuner,
                selector=selector,
                gridding=gridding,
                priority=priority,
                budget_type=budget_type,
                budget=budget,
                deadline=deadline,
                metric=metric,
                score_target=target,
                k_window=k_window,
                r_minimum=r_minimum
            )

            dataruns.append(datarun)

        for method, parts in method_parts.items():
            for part in parts:
                # if necessary, create a new datarun for each hyperpartition.
                # This setting is useful for debugging.
                if run_per_partition:
                    datarun = self.db.create_datarun(
                        dataset_id=dataset_id,
                        description=run_description,
                        tuner=tuner,
                        selector=selector,
                        gridding=gridding,
                        priority=priority,
                        budget_type=budget_type,
                        budget=budget,
                        deadline=deadline,
                        metric=metric,
                        score_target=target,
                        k_window=k_window,
                        r_minimum=r_minimum
                    )

                    dataruns.append(datarun)

                # create a new hyperpartition in the database
                self.db.create_hyperpartition(datarun_id=datarun.id,
                                              method=method,
                                              tunables=part.tunables,
                                              constants=part.constants,
                                              categoricals=part.categoricals,
                                              status=PartitionStatus.INCOMPLETE)

        dataset = self.db.get_dataset(dataset_id)
        LOGGER.info('Dataruns created. Summary:')
        LOGGER.info('\tDataset ID: {}'.format(dataset.id))
        LOGGER.info('\tTraining data: {}'.format(dataset.train_path))
        LOGGER.info('\tTest data: {}'.format(dataset.test_path))

        if run_per_partition:
            LOGGER.info('\tDatarun IDs: {}'.format(
                ', '.join(str(datarun.id) for datarun in dataruns)))

        else:
            LOGGER.info('\tDatarun ID: {}'.format(dataruns[0].id))

        LOGGER.info('\tHyperpartition selection strategy: {}'.format(dataruns[0].selector))
        LOGGER.info('\tParameter tuning strategy: {}'.format(dataruns[0].tuner))
        LOGGER.info('\tBudget: {} ({})'.format(dataruns[0].budget, dataruns[0].budget_type))

        return dataruns if run_per_partition else dataruns[0]

    def work(self, datarun_ids=None, save_files=True, choose_randomly=True,
             cloud_mode=False, total_time=None, wait=True, verbose=False):
        """
        Check the ModelHub database for unfinished dataruns, and spawn workers to
        work on them as they are added. This process will continue to run until it
        exceeds total_time or is broken with ctrl-C.

        datarun_ids (optional): list of IDs of dataruns to compute on. If None,
            this will work on all unfinished dataruns in the database.
        choose_randomly: if True, work on all highest-priority dataruns in random
            order. If False, work on them in sequential order (by ID)
        cloud_mode: if True, save processed datasets to AWS.
        total_time (optional): if set to an integer, this worker will only work for
            total_time seconds. Otherwise, it will continue working until all
            dataruns are complete (or indefinitely).
        wait: if True, once all dataruns in the database are complete, keep spinning
            and wait for new runs to be added. If False, exit once all dataruns are
            complete.
        """
        start_time = datetime.now()

        # main loop
        while True:
            # get all pending and running dataruns, or all pending/running dataruns
            # from the list we were given
            dataruns = self.db.get_dataruns(include_ids=datarun_ids, ignore_complete=True)
            if not dataruns:
                if wait:
                    LOGGER.debug('No dataruns found. Sleeping %d seconds and trying again.',
                                 ATM.LOOP_WAIT)
                    time.sleep(ATM.LOOP_WAIT)
                    continue

                else:
                    LOGGER.info('No dataruns found. Exiting.')
                    break

            # either choose a run randomly between priority, or take the run with the lowest ID
            if choose_randomly:
                run = random.choice(dataruns)
            else:
                run = sorted(dataruns, key=attrgetter('id'))[0]

            # say we've started working on this datarun, if we haven't already
            self.db.mark_datarun_running(run.id)

            LOGGER.info('Computing on datarun %d' % run.id)
            # actual work happens here
            worker = Worker(self.db, run, save_files=save_files,
                            cloud_mode=cloud_mode, aws_access_key=self.aws_access_key,
                            aws_secret_key=self.aws_secret_key, s3_bucket=self.s3_bucket,
                            s3_folder=self.s3_folder, models_dir=self.models_dir,
                            metrics_dir=self.metrics_dir, verbose_metrics=self.verbose_metrics)
            try:
                if run.budget_type == 'classifier':
                    pbar = tqdm(
                        total=run.budget,
                        ascii=True,
                        initial=run.completed_classifiers,
                        disable=not verbose
                    )

                    while run.status != RunStatus.COMPLETE:
                        worker.run_classifier()
                        run = self.db.get_datarun(run.id)
                        if verbose and run.completed_classifiers > pbar.last_print_n:
                            pbar.update(run.completed_classifiers - pbar.last_print_n)

                    pbar.close()

                elif run.budget_type == 'walltime':
                    pbar = tqdm(
                        disable=not verbose,
                        ascii=True,
                        initial=run.completed_classifiers,
                        unit=' Classifiers'
                    )

                    while run.status != RunStatus.COMPLETE:
                        worker.run_classifier()
                        run = self.db.get_datarun(run.id)  # Refresh the datarun object.
                        if verbose and run.completed_classifiers > pbar.last_print_n:
                            pbar.update(run.completed_classifiers - pbar.last_print_n)

                    pbar.close()

            except ClassifierError:
                # the exception has already been handled; just wait a sec so we
                # don't go out of control reporting errors
                LOGGER.error('Something went wrong. Sleeping %d seconds.', ATM.LOOP_WAIT)
                time.sleep(ATM.LOOP_WAIT)

            elapsed_time = (datetime.now() - start_time).total_seconds()
            if total_time is not None and elapsed_time >= total_time:
                LOGGER.info('Total run time for worker exceeded; exiting.')
                break

    def run(self, train_path, test_path=None, name=None, description=None,
            class_column='class', budget=100, budget_type='classifier', gridding=0, k_window=3,
            metric='f1', methods=['logreg', 'dt', 'knn'], r_minimum=2, run_per_partition=False,
            score_target='cv', selector='uniform', tuner='uniform', deadline=None, priority=1,
            save_files=True, choose_randomly=True, cloud_mode=False, total_time=None,
            wait=True, verbose=True):

        dataset = self.add_dataset(train_path, test_path, name, description, class_column)
        datarun = self.add_datarun(
            dataset.id,
            budget,
            budget_type,
            gridding,
            k_window,
            metric,
            methods,
            r_minimum,
            run_per_partition,
            score_target,
            priority,
            selector,
            tuner,
            deadline
        )

        if run_per_partition:
            datarun_ids = [_datarun.id for _datarun in datarun]

        else:
            datarun_ids = [datarun.id]

        if verbose:
            print('Processing dataset {}'.format(train_path))

        self.work(
            datarun_ids,
            save_files,
            choose_randomly,
            cloud_mode,
            total_time,
            False,
            verbose=verbose
        )

        dataruns = self.db.get_dataruns(
            include_ids=datarun_ids,
            ignore_complete=False,
            ignore_pending=True
        )

        if run_per_partition:
            return dataruns

        elif len(dataruns) == 1:
            return dataruns[0]

    def load_model(self, classifier_id):
        return self.db.get_classifier(classifier_id).load_model()



from __future__ import absolute_import, unicode_literals

import argparse
import os
import re
from builtins import object, str

import yaml

from atm.constants import (
    BUDGET_TYPES, CUSTOM_CLASS_REGEX, JSON_REGEX, METHODS, METRICS, SCORE_TARGETS, SELECTORS,
    SQL_DIALECTS, TIME_FMT, TUNERS)


class Config(object):
    """
    Class which stores configuration for one aspect of ATM. Subclasses of
    Config should define the list of all configurable parameters and any
    default values for those parameters other than None (in PARAMETERS and
    DEFAULTS, respectively). The object can be initialized with any number of
    keyword arguments; only kwargs that are in PARAMETERS will be used. This
    means you can (relatively) safely do things like
        args = parser.parse_args()
        conf = Config(**vars(args))
    and only relevant parameters will be set.

    Subclasses do not need to define __init__ or any other methods.
    """
    _PREFIX = None
    _CONFIG = None

    @classmethod
    def _add_prefix(cls, name):
        if cls._PREFIX:
            return '{}_{}'.format(cls._PREFIX, name)
        else:
            return name

    @classmethod
    def _get_arg(cls, args, name, use_prefix):
        class_value = getattr(cls, name)

        if use_prefix:
            name = cls._add_prefix(name)

        required = False
        if isinstance(class_value, dict):
            required = 'default' not in class_value
            default = class_value.get('default')
        elif isinstance(class_value, tuple):
            required = False
            default = class_value[1]
        else:
            required = False
            default = None

        if required and name not in args:
            raise KeyError(name)

        return args.get(name, default)

    def __init__(self, args, path=None):
        if isinstance(args, argparse.Namespace):
            args = vars(args)

        config_arg = self._CONFIG or self._PREFIX
        if not path and config_arg:
            path = args.get(config_arg + '_config')

        if path:
            with open(path, 'r') as f:
                args = yaml.load(f)
                use_prefix = False
        else:
            use_prefix = True

        for name, value in vars(self.__class__).items():
            if not name.startswith('_') and not callable(value):
                setattr(self, name, self._get_arg(args, name, use_prefix))

    @classmethod
    def get_parser(cls):
        parser = argparse.ArgumentParser(add_help=False)

        # make sure the text for these arguments is formatted correctly
        # this allows newlines in the help strings
        parser.formatter_class = argparse.RawTextHelpFormatter

        if cls._PREFIX:
            parser.add_argument('--{}-config'.format(cls._PREFIX),
                                help='path to yaml {} config file'.format(cls._PREFIX))

        for name, description in vars(cls).items():
            if not name.startswith('_') and not callable(description):
                arg_name = '--' + cls._add_prefix(name).replace('_', '-')

                if isinstance(description, dict):
                    parser.add_argument(arg_name, **description)

                elif isinstance(description, tuple):
                    description, default = description
                    parser.add_argument(arg_name, help=description, default=default)

                else:
                    parser.add_argument(arg_name, help=description)

        return parser

    def to_dict(self):
        return {
            name: value
            for name, value in vars(self).items()
            if not name.startswith('_') and not callable(value)
        }

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.to_dict())


class AWSConfig(Config):
    """ Stores configuration for AWS S3 connections """
    _PREFIX = 'aws'

    access_key = 'AWS access key'
    secret_key = 'AWS secret key'
    s3_bucket = 'AWS S3 bucket to store data'
    s3_folder = 'Folder in AWS S3 bucket in which to store data'


class DatasetConfig(Config):
    """ Stores configuration of a Dataset """
    _CONFIG = 'run'

    name = 'Given name for this dataset.'
    train_path = {
        'help': 'Path to raw training data',
        'required': True
    }
    test_path = 'Path to raw test data (if applicable)'
    description = 'Description of dataset'
    class_column = ('Name of the class column in the input data', 'class')


class SQLConfig(Config):
    """ Stores configuration for SQL database setup & connection """
    _PREFIX = 'sql'

    dialect = {
        'help': 'Dialect of SQL to use',
        'default': 'sqlite',
        'choices': SQL_DIALECTS
    }
    database = ('Name of, or path to, SQL database', 'atm.db')
    username = 'Username for SQL database'
    password = 'Password for SQL database'
    host = 'Hostname for database machine'
    port = 'Port used to connect to database'
    query = 'Specify extra login details'


class LogConfig(Config):
    models_dir = ('Directory where computed models will be saved', 'models')
    metrics_dir = ('Directory where model metrics will be saved', 'metrics')
    verbose_metrics = {
        'help': (
            'If set, compute full ROC and PR curves and '
            'per-label metrics for each classifier'
        ),
        'action': 'store_true',
        'default': False
    }


def option_or_path(options, regex=CUSTOM_CLASS_REGEX):
    def type_check(s):
        # first, check whether the argument is one of the preconfigured options
        if s in list(options):
            return s

        # otherwise, check it against the regex, and try to pull out a path to a
        # real file. The regex must extract the path to the file as groups()[0].
        match = re.match(regex, s)
        if match and os.path.isfile(match.groups()[0]):
            return s

        # if both of those fail, there's something wrong
        raise argparse.ArgumentTypeError('{} is not a valid option or path!'.format(s))

    return type_check


class RunConfig(Config):
    """Stores configuration for Dataset and Datarun setup."""
    _CONFIG = 'run'

    dataset_id = {
        'help': 'ID of dataset, if it is already in the database',
        'type': int
    }

    run_per_partition = {
        'help': 'if true, generate a new datarun for each hyperpartition',
        'default': False,
        'action': 'store_true',
    }

    # Method options:
    #   logreg - logistic regression
    #   svm    - support vector machine
    #   sgd    - linear classifier with stochastic gradient descent
    #   dt     - decision tree
    #   et     - extra trees
    #   rf     - random forest
    #   gnb    - gaussian naive bayes
    #   mnb    - multinomial naive bayes
    #   bnb    - bernoulli naive bayes
    #   gp     - gaussian process
    #   pa     - passive aggressive
    #   knn    - K nearest neighbors
    #   mlp    - multi-layer perceptron
    #
    # Notes:
    # - Support vector machines (svm) can take a long time to train. It's not an
    #   error, it's just part of what happens when the method happens to explore
    #   a crappy set of parameters on a powerful algo like this.
    # - Stochastic gradient descent (sgd) can sometimes fail on certain
    #   parameter settings as well. Don't worry, they train SUPER fast, and the
    #   worker.py will simply log the error and continue.
    methods = {
        'help': (
            'Method or list of methods to use for '
            'classification. Each method can either be one of the '
            'pre-defined method codes listed below or a path to a '
            'JSON file defining a custom method.\n\nOptions: [{}]'
        ).format(', '.join(str(s) for s in METHODS.keys())),
        'default': ['logreg', 'dt', 'knn'],
        'type': option_or_path(METHODS.keys(), JSON_REGEX),
        'nargs': '+'
    }

    priority = {
        'help': 'Priority of the datarun (higher = more important',
        'default': 1,
        'type': int
    }
    budget_type = {
        'help': 'Type of budget to use',
        'default': 'classifier',
        'choices': BUDGET_TYPES,
    }
    budget = {
        'help': 'Value of the budget, either in classifiers or minutes',
        'default': 100,
        'type': int,
    }
    deadline = (
        'Deadline for datarun completion. If provided, this '
        'overrides the configured walltime budget.\nFormat: {}'
    ).format(TIME_FMT.replace('%', '%%'))

    # Which field to use to judge performance, for the sake of AutoML
    # options:
    #   f1        - F1 score (harmonic mean of precision and recall)
    #   roc_auc   - area under the Receiver Operating Characteristic curve
    #   accuracy  - percent correct
    #   cohen_kappa     - measures accuracy, but controls for chance of guessing
    #                     correctly
    #   rank_accuracy   - multiclass only: percent of examples for which the true
    #                     label is in the top 1/3 most likely predicted labels
    #   ap        - average precision: nearly identical to area under
    #               precision/recall curve.
    #   mcc       - matthews correlation coefficient: good for unbalanced classes
    #
    # f1 and roc_auc may be appended with _micro or _macro to use with
    # multiclass problems.
    metric = {
        'help': (
            'Metric by which ATM should evaluate classifiers. '
            'The metric function specified here will be used to '
            'compute the "judgment metric" for each classifier.'
        ),
        'default': 'f1',
        'choices': METRICS,
    }

    # Which data to use for computing judgment score
    #   cv   - cross-validated performance on training data
    #   test - performance on test data
    #   mu_sigma - lower confidence bound on cv score
    score_target = {
        'help': (
            'Determines which judgment metric will be used to '
            'search the hyperparameter space. "cv" will use the mean '
            'cross-validated performance, "test" will use the '
            'performance on a test dataset, and "mu_sigma" will use '
            'the lower confidence bound on the CV performance.'
        ),
        'default': 'cv',
        'choices': SCORE_TARGETS
    }

    #   AutoML Arguments  ######################################################
    # ##########################################################################

    # hyperparameter selection strategy
    # How should ATM sample hyperparameters from a given hyperpartition?
    #    uniform  - pick randomly! (baseline)
    #    gp       - vanilla Gaussian Process
    #    gp_ei    - Gaussian Process expected improvement criterion
    #    gp_eivel - Gaussian Process expected improvement, with randomness added
    #               in based on velocity of improvement
    #   path to custom tuner, defined in python
    tuner = {
        'help': (
            'Type of BTB tuner to use. Can either be one of the pre-configured '
            'tuners listed below or a path to a custom tuner in the form '
            '"/path/to/tuner.py:ClassName".\n\nOptions: [{}]'
        ).format(', '.join(str(s) for s in TUNERS.keys())),
        'default': 'uniform',
        'type': option_or_path(TUNERS.keys())
    }

    # How should ATM select a particular hyperpartition from the set of all
    # possible hyperpartitions?
    # Options:
    #   uniform      - pick randomly
    #   ucb1         - UCB1 multi-armed bandit
    #   bestk        - MAB using only the best K runs in each hyperpartition
    #   bestkvel     - MAB with velocity of best K runs
    #   purebestkvel - always return hyperpartition with highest velocity
    #   recentk      - MAB with most recent K runs
    #   recentkvel   - MAB with velocity of most recent K runs
    #   hieralg      - hierarchical MAB: choose a classifier first, then choose
    #                  a partition
    #   path to custom selector, defined in python
    selector = {
        'help': (
            'Type of BTB selector to use. Can either be one of the pre-configured '
            'selectors listed below or a path to a custom tuner in the form '
            '"/path/to/selector.py:ClassName".\n\nOptions: [{}]'
        ).format(', '.join(str(s) for s in SELECTORS.keys())),
        'default': 'uniform',
        'type': option_or_path(SELECTORS.keys())
    }

    # r_minimum is the number of random runs performed in each hyperpartition before
    # allowing bayesian opt to select parameters. Consult the thesis to
    # understand what those mean, but essentially:
    #
    #  if (num_classifiers_trained_in_hyperpartition >= r_minimum)
    #    # train using sample criteria
    #  else
    #    # train using uniform (baseline)
    r_minimum = {
        'help': 'number of random runs to perform before tuning can occur',
        'default': 2,
        'type': int
    }

    # k is number that xxx-k methods use. It is similar to r_minimum, except it is
    # called k_window and determines how much "history" ATM considers for certain
    # partition selection logics.
    k_window = {
        'help': 'number of previous scores considered by -k selector methods',
        'default': 3,
        'type': int
    }

    # gridding determines whether or not sample selection will happen on a grid.
    # If any positive integer, a grid with `gridding` points on each axis is
    # established, and hyperparameter vectors are sampled from this finite
    # space. If 0 (or blank), hyperparameters are sampled from continuous
    # space, and there is no limit to the number of hyperparameter vectors that
    # may be tried.
    gridding = {
        'help': 'gridding factor (0: no gridding)',
        'default': 0,
        'type': int
    }



import random

from btb.selection.selector import Selector


class MySelector(Selector):
    def select(self, choice_scores):
        """ Select a choice uniformly at random.  """
        return self.choices[random.randint(0, len(self.choices) - 1)]



"""Auto Tune Models
A multi-user, multi-data AutoML framework.
"""
from __future__ import absolute_import, unicode_literals

import os

from atm.classifier import Model
from atm.core import ATM

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.2.0'

# this defines which modules will be imported by "from atm import *"
__all__ = ['ATM', 'Model', 'config', 'constants', 'data', 'database',
           'method', 'metrics', 'models', 'utilities', 'worker']

# Get the path of the project root, so that the rest of the project can
# reference files relative to there.
PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))



import glob
import logging
import os
import shutil

import boto3
import pandas as pd
import requests
from botocore.exceptions import ClientError

LOGGER = logging.getLogger('atm')


def copy_files(pattern, source, target=None):
    if isinstance(source, (list, tuple)):
        source = os.path.join(*source)

    if target is None:
        target = source

    source_dir = os.path.join(os.path.dirname(__file__), source)
    target_dir = os.path.join(os.getcwd(), target)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    datasets = dict()

    for source_file in glob.glob(os.path.join(source_dir, pattern)):
        file_name = os.path.basename(source_file)
        target_file = os.path.join(target_dir, file_name)
        print('Generating file {}'.format(target_file))
        shutil.copy(source_file, target_file)
        datasets[file_name.replace('.csv', '')] = target_file

    return datasets


def get_demos(args=None):
    return copy_files('*.csv', 'demos', 'demos')


def download_from_s3(path, local_path, aws_access_key=None, aws_secret_key=None, **kwargs):

    client = boto3.client(
        's3',
        aws_access_key_id=aws_access_key,
        aws_secret_access_key=aws_secret_key,
    )

    bucket = path.split('/')[2]
    file_to_download = path.replace('s3://{}/'.format(bucket), '')

    try:
        LOGGER.info('Downloading {}'.format(path))
        client.download_file(bucket, file_to_download, local_path)

        return local_path

    except ClientError as e:
        LOGGER.error('An error occurred trying to download from AWS3.'
                     'The following error has been returned: {}'.format(e))


def download_from_url(url, local_path, **kwargs):

    data = requests.get(url).text
    with open(local_path, 'wb') as outfile:
        outfile.write(data.encode())

    LOGGER.info('File saved at {}'.format(local_path))

    return local_path


DOWNLOADERS = {
    's3': download_from_s3,
    'http': download_from_url,
    'https': download_from_url,
}


def download(path, local_path, **kwargs):
    protocol = path.split(':', 1)[0]
    downloader = DOWNLOADERS.get(protocol)

    if not downloader:
        raise ValueError('Unknown protocol: {}'.format(protocol))

    return downloader(path, local_path, **kwargs)


def get_local_path(name, path, aws_access_key=None, aws_secret_key=None):

    if os.path.isfile(path):
        return path

    cwd = os.getcwd()
    data_path = os.path.join(cwd, 'data')

    if not name.endswith('csv'):
        name = name + '.csv'

    local_path = os.path.join(data_path, name)

    if os.path.isfile(local_path):
        return local_path

    if not os.path.isfile(local_path):
        if not os.path.exists(data_path):
            os.makedirs(data_path)

        download(path, local_path, aws_access_key=aws_access_key, aws_secret_key=aws_secret_key)
        return local_path


def load_data(name, path, aws_access_key=None, aws_secret_key=None):
    local_path = get_local_path(
        name, path, aws_access_key=aws_access_key, aws_secret_key=aws_secret_key)

    # load data as a Pandas dataframe
    return pd.read_csv(local_path).dropna(how='any')



from __future__ import division, unicode_literals

from builtins import object

import numpy as np
from sklearn.preprocessing import LabelEncoder, OneHotEncoder


class DataEncoder(object):
    def __init__(self, class_column='class', feature_columns=None):
        self.class_column = class_column
        self.feature_columns = feature_columns

        # these will be trained with fit_encoders()
        self.column_encoders = {}
        self.feature_encoder = None
        self.label_encoder = None

    def fit(self, data):
        """
        Fit one-hot encoders for categorical features and an integer encoder for
        the label. These can be used later to transform raw data into a form
        that ATM can work with.

        data: pd.DataFrame of unprocessed data
        """
        if self.class_column not in data.columns:
            raise KeyError('Class column "%s" not found in dataset!' %
                           self.class_column)

        self.categorical_columns = []
        if self.feature_columns is None:
            X = data.drop([self.class_column], axis=1)
            self.feature_columns = X.columns
        else:
            X = data[self.feature_columns]

        # encode categorical columns, leave ordinal values alone
        for column in X.columns:
            if X[column].dtype == 'object':
                # save the indices of categorical columns for one-hot encoding
                self.categorical_columns.append(X.columns.get_loc(column))

                # encode each feature as an integer in range(unique_vals)
                le = LabelEncoder()
                X[column] = le.fit_transform(X[column])
                self.column_encoders[column] = le

        # One-hot encode the whole feature matrix.
        # Set sparse to False so that we can test for NaNs in the output
        if self.categorical_columns:
            self.feature_encoder = OneHotEncoder(
                categorical_features=self.categorical_columns,
                sparse=False,
                handle_unknown='ignore'
            )
            self.feature_encoder.fit(X)

        # Train an encoder for the label as well
        labels = np.array(data[[self.class_column]])
        self.label_encoder = LabelEncoder()
        self.label_encoder.fit(labels)

    def transform(self, data):
        """
        Convert a DataFrame of labeled data to a feature matrix in the form
        that ATM can use.
        """
        if self.class_column in data:
            # pull labels into a separate series and transform them to integers
            labels = np.array(data[[self.class_column]])
            y = self.label_encoder.transform(labels)
            # drop the label column and transform the remaining features
        else:
            y = None

        X = data[self.feature_columns]

        # one-hot encode the categorical X
        if self.categorical_columns:

            # encode each categorical feature as an integer
            for column, encoder in list(self.column_encoders.items()):
                X[column] = encoder.transform(X[column])

            X = self.feature_encoder.transform(X)

        else:
            X = X.values

        return X, y

    def fit_transform(self, data):
        """ Process data into a form that ATM can use. """
        self.fit(data)
        return self.transform(data)



from __future__ import absolute_import, unicode_literals

import base64
import hashlib
import json
import logging
import os
import pickle
from builtins import str

import numpy as np

from atm.compat import getargs

logger = logging.getLogger('atm')


def hash_dict(dictionary, ignored_keys=None):
    """
    Hash a python dictionary to a hexadecimal string.
    http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
    """
    dictionary = dict(dictionary)  # copy dictionary
    for key in (ignored_keys or []):
        del dictionary[key]
    return hashlib.md5(repr(sorted(dictionary.items())).encode('utf8')).hexdigest()


def hash_nested_tuple(tup):
    """ Hash a nested tuple to hexadecimal """
    return hashlib.md5(repr(sorted(tup)).encode('utf8')).hexdigest()


def hash_string(s):
    """ Hash a string to hexadecimal """
    return hashlib.md5(str(s).encode('utf8')).hexdigest()


def ensure_directory(directory):
    """ Create directory if it doesn't exist. """
    if not os.path.exists(directory):
        os.makedirs(directory)


def object_to_base_64(obj):
    """ Pickle and base64-encode an object. """
    pickled = pickle.dumps(obj)
    return base64.b64encode(pickled)


def base_64_to_object(b64str):
    """
    Inverse of object_to_base_64.
    Decode base64-encoded string and then unpickle it.
    """
    decoded = base64.b64decode(b64str)
    return pickle.loads(decoded)


def obj_has_method(obj, method):
    """http://stackoverflow.com/questions/34439/finding-what-methods-an-object-has"""
    return hasattr(obj, method) and callable(getattr(obj, method))


# Converting hyperparameters to and from BTB-compatible formats

def update_params(params, categoricals, constants):
    """
    Update params with categoricals and constants for the fitting proces.

    params: params proposed by the tuner

    Examples of the format for SVM sigmoid hyperpartition:

    categoricals = (('kernel', 'poly'),
                    ('probability', True),
                    ('_scale', True))

    constants = [('cache_size', 15000)]
    """
    for key, value in categoricals + constants:
        params[key] = value

    return params


def get_instance(class_, **kwargs):
    """Instantiate an instance of the given class with unused kwargs

    Args:
        class_ (type): class to instantiate
        **kwargs: keyword arguments to specific selector class

    Returns:
        instance of specific class with the args that accepts.
    """
    init_args = getargs(class_.__init__)
    relevant_kwargs = {
        k: kwargs[k]
        for k in kwargs
        if k in init_args
    }

    return class_(**relevant_kwargs)


def params_to_vectors(params, tunables):
    """
    Converts a list of parameter vectors (with metadata) into a numpy array
    ready for BTB tuning.

    Args:
        params: list of hyperparameter vectors. Each vector is a dict mapping
            the names of parameters to those parameters' values.
        tunables: list of HyperParameter metadata structures describing all
            the optimizable hyperparameters that should be in each vector. e.g.

        tunables = [('C',  HyperParameter(type='float_exp', range=(1e-5, 1e5))),
                    ('degree',  HyperParameter('int', (2, 4))),
                    ('gamma',  HyperParameter('float_exp', (1e-05, 1e5)))]

    Returns:
        vectors: np.array of parameter vectors ready to be optimized by a
            Gaussian Process (or what have you).
            vectors.shape = (len(params), len(tunables))
    """
    # make sure params is iterable
    if not isinstance(params, (list, np.ndarray)):
        params = [params]

    keys = [k[0] for k in tunables]
    vectors = np.zeros((len(params), len(keys)))
    for i, p in enumerate(params):
        for j, k in enumerate(keys):
            vectors[i, j] = p[k]

    return vectors


# Serializing and deserializing data on disk

def make_save_path(dir, classifier, suffix):
    """
    Generate the base save path for a classifier's model and metrics files,
    based on the classifier's dataset name and hyperparameters.
    """
    run_name = "".join([c for c in classifier.datarun.dataset.name
                        if c.isalnum() or c in (' ', '-', '_')]).rstrip()
    params_hash = hash_dict(classifier.hyperparameter_values)[:8]
    filename = "%s-%s.%s" % (run_name, params_hash, suffix)
    return os.path.join(dir, filename)


def save_model(classifier, models_dir, model):
    """
    Save a serialized version of a Model object for a particular classifier.
    The object will be stored at a path generated from the classifier's
    attributes.
    """
    path = make_save_path(models_dir, classifier, 'model')
    logger.info('Saving model in: %s' % path)
    with open(path, 'wb') as f:
        pickle.dump(model, f, protocol=pickle.HIGHEST_PROTOCOL)
    return path


def save_metrics(classifier, metrics_dir, metrics):
    """
    Save a JSON-serialized version of a set of performance metrics for a
    particular classifier. The metrics will be stored at a path generated from
    the classifier's attributes.
    """
    path = make_save_path(metrics_dir, classifier, 'metric')
    logger.info('Saving metrics in: %s' % path)
    with open(path, 'w') as f:
        json.dump(metrics, f)
    return path


def load_model(classifier, models_dir):
    """ Load the Model object for a particular classifier """
    path = make_save_path(models_dir, classifier, 'model')
    with open(path, 'rb') as f:
        return pickle.load(f)


def load_metrics(classifier, metrics_dir):
    """ Load the performance metrics for a particular classifier """
    path = make_save_path(metrics_dir, classifier, 'metric')
    with open(path) as f:
        return json.load(f)



import os

import pytest

from atm.core import ATM
from atm.data import get_local_path
from atm.database import Database

DB_PATH = '/tmp/atm.db'
DATA_URL = 'https://atm-data.s3.amazonaws.com/'
DB_CACHE_PATH = os.path.join(os.path.dirname(__file__), 'modelhub')

METHOD_HYPERPARTS = {
    'logreg': 6,
    'svm': 4,
    'sgd': 24,
    'dt': 2,
    'et': 2,
    'rf': 2,
    'gnb': 1,
    'mnb': 1,
    'bnb': 1,
    'gp': 8,
    'pa': 4,
    'knn': 24,
    'mlp': 60,
}


@pytest.fixture
def db():
    if os.path.exists(DB_PATH):
        os.remove(DB_PATH)

    db = Database(dialect='sqlite', database=DB_PATH)

    # load cached ModelHub state. This database snapshot has one dataset
    # (pollution.csv) and two dataruns, one complete and one with 33/100
    # classifiers finished.
    db.from_csv(DB_CACHE_PATH)

    return db


@pytest.fixture
def dataset(db):
    return db.get_dataset(1)


def test_create_dataset(db):
    train_url = DATA_URL + 'pollution_1_train.csv'
    test_url = DATA_URL + 'pollution_1_test.csv'

    train_path_local = get_local_path('pollution_test.csv', train_url, None)
    if os.path.exists(train_path_local):
        os.remove(train_path_local)

    test_path_local = get_local_path('pollution_test_test.csv', test_url, None)
    if os.path.exists(test_path_local):
        os.remove(test_path_local)

    atm = ATM(database=DB_PATH)

    dataset = atm.add_dataset(
        name='pollution_test',
        train_path=train_url,
        test_path=test_url,
        description='test',
        class_column='class'
    )
    dataset = db.get_dataset(dataset.id)

    train, test = dataset.load()  # This will create the test_path_local

    assert os.path.exists(train_path_local)
    assert os.path.exists(test_path_local)

    assert dataset.train_path == train_url
    assert dataset.test_path == test_url
    assert dataset.description == 'test'
    assert dataset.class_column == 'class'
    assert dataset.n_examples == 40
    assert dataset.d_features == 16
    assert dataset.k_classes == 2
    assert dataset.majority >= 0.5

    # remove test dataset
    if os.path.exists(train_path_local):
        os.remove(train_path_local)

    if os.path.exists(test_path_local):
        os.remove(test_path_local)



"""
.. module:: wrapper
   :synopsis: Model around classification method.

"""
from __future__ import absolute_import, division, unicode_literals

import logging
import os
import pickle
import re
import time
from builtins import object
from collections import defaultdict
from importlib import import_module

import numpy as np
import pandas as pd
from sklearn import decomposition
from sklearn.gaussian_process.kernels import (
    RBF, ConstantKernel, ExpSineSquared, Matern, RationalQuadratic)
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from atm.constants import Metrics
from atm.encoder import DataEncoder
from atm.method import Method
from atm.metrics import cross_validate_pipeline, test_pipeline

# load the library-wide logger
logger = logging.getLogger('atm')


class Model(object):
    """
    This class contains everything needed to run an end-to-end ATM classifier
    pipeline. It is initialized with a set of parameters and trained like a
    normal sklearn model. This class can be pickled and saved to disk, then
    unpickled outside of ATM and used to classify new datasets.
    """
    # these are special keys that are used for general purpose
    # things like scaling, normalization, PCA, etc
    SCALE = "_scale"
    WHITEN = "_whiten"
    MINMAX = "_scale_minmax"
    PCA = "_pca"
    PCA_DIMS = "_pca_dimensions"

    # list of all such keys
    ATM_KEYS = [SCALE, WHITEN, MINMAX, PCA, PCA_DIMS]

    # number of folds for cross-validation (arbitrary, for speed)
    N_FOLDS = 5

    def __init__(self, method, params, judgment_metric, class_column,
                 testing_ratio=0.3, verbose_metrics=False):
        """
        Parameters:
            method: the short method code (as defined in constants.py) or path
                to method json
            judgment_metric: string that indicates which metric should be
                optimized for.
            params: parameters passed to the sklearn classifier constructor
            class_: sklearn classifier class
        """
        # configuration & database
        self.method = method
        self.params = params
        self.judgment_metric = judgment_metric
        self.class_column = class_column
        self.testing_ratio = testing_ratio
        self.verbose_metrics = verbose_metrics

        # load the classifier method's class
        path = Method(method).class_path.split('.')
        mod_str, cls_str = '.'.join(path[:-1]), path[-1]
        mod = import_module(mod_str)
        self.class_ = getattr(mod, cls_str)

        # pipelining
        self.pipeline = None

        # persistent random state
        self.random_state = np.random.randint(1e7)

    def make_pipeline(self):
        """
        Makes the classifier as well as scaling or dimension reduction steps.
        """
        # create a list of steps, starting with the data encoder
        steps = []

        # create a classifier with specified parameters
        hyperparameters = {k: v for k, v in list(self.params.items())
                           if k not in Model.ATM_KEYS}
        atm_params = {k: v for k, v in list(self.params.items())
                      if k in Model.ATM_KEYS}

        # do special conversions
        hyperparameters = self.special_conversions(hyperparameters)
        classifier = self.class_(**hyperparameters)

        if Model.PCA in atm_params and atm_params[Model.PCA]:
            whiten = (Model.WHITEN in atm_params and atm_params[Model.WHITEN])
            pca_dims = atm_params[Model.PCA_DIMS]
            # PCA dimension in atm_params is a float reprsenting percentages of
            # features to use
            if pca_dims < 1:
                dimensions = int(pca_dims * float(self.num_features))
                logger.info("Using PCA to reduce %d features to %d dimensions"
                            % (self.num_features, dimensions))
                pca = decomposition.PCA(n_components=dimensions, whiten=whiten)
                steps.append(('pca', pca))

        # should we scale the data?
        if atm_params.get(Model.SCALE):
            steps.append(('standard_scale', StandardScaler()))
        elif self.params.get(Model.MINMAX):
            steps.append(('minmax_scale', MinMaxScaler()))

        # add the classifier as the final step in the pipeline
        steps.append((self.method, classifier))
        self.pipeline = Pipeline(steps)

    def cross_validate(self, X, y):
        # TODO: this is hacky. See https://github.com/HDI-Project/ATM/issues/48
        binary = self.num_classes == 2
        kwargs = {}
        if self.verbose_metrics:
            kwargs['include_curves'] = True
            if not binary:
                kwargs['include_per_class'] = True

        df, cv_scores = cross_validate_pipeline(pipeline=self.pipeline,
                                                X=X, y=y, binary=binary,
                                                n_folds=self.N_FOLDS, **kwargs)

        self.cv_judgment_metric = np.mean(df[self.judgment_metric])
        self.cv_judgment_metric_stdev = np.std(df[self.judgment_metric])
        cv_stdev = (2 * self.cv_judgment_metric_stdev)
        self.mu_sigma_judgment_metric = self.cv_judgment_metric - cv_stdev

        return cv_scores

    def test_final_model(self, X, y):
        """
        Test the (already trained) model pipeline on the provided test data
        (X and y). Store the test judgment metric and return the rest of the
        metrics as a hierarchical dictionary.
        """
        # time the prediction
        start_time = time.time()
        total = time.time() - start_time
        self.avg_predict_time = total / float(len(y))

        # TODO: this is hacky. See https://github.com/HDI-Project/ATM/issues/48
        binary = self.num_classes == 2
        kwargs = {}
        if self.verbose_metrics:
            kwargs['include_curves'] = True
            if not binary:
                kwargs['include_per_class'] = True

        # compute the actual test scores!
        test_scores = test_pipeline(self.pipeline, X, y, binary, **kwargs)

        # save meta-metrics
        self.test_judgment_metric = test_scores.get(self.judgment_metric)

        return test_scores

    def train_test(self, dataset):

        self.num_classes = dataset.k_classes
        self.num_features = dataset.d_features

        # if necessary, cast judgment metric into its binary/multiary equivalent
        if self.num_classes == 2:
            if self.judgment_metric in [Metrics.F1_MICRO, Metrics.F1_MACRO]:
                self.judgment_metric = Metrics.F1
            elif self.judgment_metric in [Metrics.ROC_AUC_MICRO,
                                          Metrics.ROC_AUC_MACRO]:
                self.judgment_metric = Metrics.ROC_AUC
        else:
            if self.judgment_metric == Metrics.F1:
                self.judgment_metric = Metrics.F1_MACRO
            elif self.judgment_metric == Metrics.ROC_AUC:
                self.judgment_metric = Metrics.ROC_AUC_MACRO

        # load training data
        train_data, test_data = dataset.load(self.testing_ratio, self.random_state)

        # extract feature matrix and labels from raw data
        self.encoder = DataEncoder(class_column=self.class_column)
        self.encoder.fit(train_data)
        X_train, y_train = self.encoder.transform(train_data)
        X_test, y_test = self.encoder.transform(test_data)

        # create and cross-validate pipeline
        self.make_pipeline()
        cv_scores = self.cross_validate(X_train, y_train)

        # train and test the final model
        self.pipeline.fit(X_train, y_train)
        test_scores = self.test_final_model(X_test, y_test)
        return {'cv': cv_scores, 'test': test_scores}

    def predict(self, data):
        """
        Use the trained encoder and pipeline to transform training data into
        predicted labels

        data: pd.DataFrame of data for which to predict classes

        returns: pd.Series populated with predictions, with index matching data.
        """
        X, _ = self.encoder.transform(data)
        predictions = self.pipeline.predict(X)
        labels = self.encoder.label_encoder.inverse_transform(predictions)
        labels = pd.Series(labels, index=data.index)
        return labels

    def special_conversions(self, params):
        """
        TODO: replace this logic with something better
        """
        # create list parameters
        lists = defaultdict(list)
        element_regex = re.compile(r'(.*)\[(\d)\]')
        for name, param in list(params.items()):
            # look for variables of the form "param_name[1]"
            match = element_regex.match(name)
            if match:
                # name of the list parameter
                lname = match.groups()[0]
                # index of the list item
                index = int(match.groups()[1])
                lists[lname].append((index, param))

                # drop the element parameter from our list
                del params[name]

        for lname, items in list(lists.items()):
            # drop the list size parameter
            del params['len(%s)' % lname]

            # sort the list by index
            params[lname] = [val for idx, val in sorted(items)]

        # Gaussian process classifier
        if self.method == "gp":
            if params["kernel"] == "constant":
                params["kernel"] = ConstantKernel()
            elif params["kernel"] == "rbf":
                params["kernel"] = RBF()
            elif params["kernel"] == "matern":
                params["kernel"] = Matern(nu=params["nu"])
                del params["nu"]
            elif params["kernel"] == "rational_quadratic":
                params["kernel"] = RationalQuadratic(length_scale=params["length_scale"],
                                                     alpha=params["alpha"])
                del params["length_scale"]
                del params["alpha"]
            elif params["kernel"] == "exp_sine_squared":
                params["kernel"] = ExpSineSquared(length_scale=params["length_scale"],
                                                  periodicity=params["periodicity"])
                del params["length_scale"]
                del params["periodicity"]

        # return the updated parameter vector
        return params

    @classmethod
    def load(cls, path):
        """Loads a saved instance from a path."""

        with open(path, 'rb') as classifier:
            return pickle.load(classifier)

    def save(self, path, force=False):
        if os.path.exists(path) and not force:
            print('The indicated path already exists. Use `force=True` to overwrite.')

        base_path = os.path.dirname(path)
        if not os.path.exists(base_path):
            os.makedirs(base_path)

        with open(path, 'wb') as pickle_file:
            pickle.dump(self, pickle_file)

        print("Model saved as {}".format(path))



from flask import Flask, jsonify, redirect, request
from flask_restless_swagger import SwagAPIManager as APIManager
from flask_sqlalchemy import SQLAlchemy

from atm.api.utils import auto_abort, make_absolute
from atm.config import RunConfig


def create_app(atm, debug=False):
    db = atm.db
    app = Flask(__name__)
    app.config['DEBUG'] = debug
    app.config['SQLALCHEMY_DATABASE_URI'] = make_absolute(db.engine.url)

    # Create the Flask-Restless API manager.
    manager = APIManager(app, flask_sqlalchemy_db=SQLAlchemy(app))

    @app.route('/api/run', methods=['POST'])
    @auto_abort((KeyError, ValueError))
    def atm_run():
        data = request.json
        run_conf = RunConfig(data)

        dataruns = atm.create_dataruns(run_conf)

        response = {
            'status': 200,
            'datarun_ids': [datarun.id for datarun in dataruns]
        }

        return jsonify(response)

    @app.route('/')
    def swagger():
        return redirect('/static/swagger/swagger-ui/index.html')

    # Create API endpoints, which will be available at /api/<tablename> by
    # default. Allowed HTTP methods can be specified as well.
    manager.create_api(db.Dataset, methods=['GET', 'POST'])
    manager.create_api(db.Datarun, methods=['GET'])
    manager.create_api(db.Hyperpartition, methods=['GET'])
    manager.create_api(db.Classifier, methods=['GET'])

    return app



import datetime
import os
import random

import numpy as np
import pytest
from btb.selection import BestKVelocity
from btb.selection.selector import Selector
from btb.tuning import GP
from btb.tuning.tuner import BaseTuner
from mock import ANY, Mock, patch

from atm.classifier import Model
from atm.config import DatasetConfig, RunConfig
from atm.constants import METRICS_BINARY, TIME_FMT
from atm.core import ATM
from atm.database import Database, db_session
from atm.utilities import load_metrics, load_model
from atm.worker import ClassifierError, Worker

CURDIR = os.path.dirname(__file__)
DEMOS_PATH = os.path.realpath(os.path.join(CURDIR, os.pardir, 'atm', 'demos'))
POLLUTION_PATH = os.path.join(DEMOS_PATH, 'pollution.csv')
DB_CACHE_PATH = os.path.join(CURDIR, 'modelhub')
DB_PATH = '/tmp/atm.db'
METRIC_DIR = '/tmp/metrics/'
MODEL_DIR = '/tmp/models/'

DATASET_ID = 1
DATARUN_ID = 2
HYPERPART_ID = 34
DT_PARAMS = {
    'criterion': 'gini',
    'max_features': 0.5,
    'max_depth': 3,
    'min_samples_split': 2,
    'min_samples_leaf': 1
}

TESTS_ROOT = os.path.dirname(__file__)


# helper class to allow fuzzy arg matching
class StringWith(object):
    def __init__(self, match):
        self.match = match

    def __eq__(self, other):
        return self.match in other


# helper class to allow incomplete object matching
class ObjWithAttrs(object):
    def __init__(self, **kwargs):
        self.attrs = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __eq__(self, other):
        return all([getattr(other, k) == v for k, v in self.attrs.items()])

    def __repr__(self):
        return '<%s>' % ', '.join(['%s=%s' % i for i in self.attrs.items()])


@pytest.fixture
def db():
    os.remove(DB_PATH)
    db = Database(dialect='sqlite', database=DB_PATH)
    # load cached ModelHub state. This database snapshot has one dataset
    # (pollution.csv) and two dataruns, one complete and one with 33/100
    # classifiers finished.
    db.from_csv(DB_CACHE_PATH)
    return db


@pytest.fixture
def dataset(db):
    # return the only dataset
    return db.get_dataset(DATASET_ID)


@pytest.fixture
def datarun(db):
    # return the unfinished datarun
    return db.get_datarun(DATARUN_ID)


@pytest.fixture
def hyperpartition(db):
    # return a decision tree hyperpartition matching the static params above
    return db.get_hyperpartition(HYPERPART_ID)


@pytest.fixture
def model(dataset):
    model = Model(method='dt', params=DT_PARAMS,
                  judgment_metric='roc_auc',
                  class_column=dataset.class_column)
    model.train_test(dataset)
    return model


@pytest.fixture
def metrics():
    cv_mets = [{k: random.random() for k in METRICS_BINARY} for i in range(5)]
    test_mets = {k: random.random() for k in METRICS_BINARY}
    return {'cv': cv_mets, 'test': test_mets}


@pytest.fixture
def worker(db, datarun):
    return Worker(db, datarun)


def get_new_worker(**kwargs):
    kwargs['dataset_id'] = kwargs.get('dataset_id', None)
    kwargs['methods'] = kwargs.get('methods', ['logreg', 'dt'])
    run_conf = RunConfig(kwargs)

    kwargs['train_path'] = POLLUTION_PATH
    dataset_conf = DatasetConfig(kwargs)

    db = Database(dialect='sqlite', database=DB_PATH)
    atm = ATM(dialect='sqlite', database=DB_PATH)

    dataset = atm.add_dataset(**dataset_conf.to_dict())
    run_conf.dataset_id = dataset.id
    datarun = atm.add_datarun(**run_conf.to_dict())

    return Worker(db, datarun)


def test_load_selector_and_tuner(db, dataset):
    worker = get_new_worker(selector='bestkvel', k_window=7, tuner='gp')
    assert isinstance(worker.selector, BestKVelocity)
    assert len(worker.selector.choices) == 8
    assert worker.selector.k == 7
    assert worker.Tuner == GP


def test_load_custom_selector_and_tuner(db, dataset):
    tuner_path = os.path.join(TESTS_ROOT, 'utilities', 'mytuner.py')
    selector_path = os.path.join(TESTS_ROOT, 'utilities', 'myselector.py')
    worker = get_new_worker(selector=selector_path + ':MySelector',
                            tuner=tuner_path + ':MyTuner')
    assert isinstance(worker.selector, Selector)
    assert issubclass(worker.Tuner, BaseTuner)


def test_select_hyperpartition(worker):
    """
    This won't test that BTB is working correctly, just that the ATM-BTB
    connection is working.
    """
    worker.db.get_hyperpartitions = Mock(return_value=[Mock(id=1)])
    clf_mock = Mock(hyperpartition_id=1, cv_judgment_metric=0.5)
    worker.db.get_classifiers = Mock(return_value=[clf_mock])
    worker.selector.select = Mock(return_value=1)

    hp = worker.select_hyperpartition()

    worker.selector.select.assert_called_with({1: [0.5]})
    assert hp.id == 1


def test_tune_hyperparameters(worker, hyperpartition):
    """
    This won't test that BTB is working correctly, just that the ATM-BTB
    connection is working.
    """
    mock_tuner = Mock()
    worker.Tuner = Mock(return_value=mock_tuner)

    with patch('atm.worker.update_params') as update_params_mock:
        worker.tune_hyperparameters(hyperpartition)

        update_params_mock.assert_called_once_with(
            params=mock_tuner.propose.return_value,
            categoricals=hyperpartition.categoricals,
            constants=hyperpartition.constants
        )

    mock_tuner.propose.assert_called()


def test_test_classifier(db, dataset):
    metric = 'roc_auc'
    worker = get_new_worker(metric=metric, score_target='mu_sigma')

    model, metrics = worker.test_classifier(method='dt', params=DT_PARAMS)
    judge_mets = [m[metric] for m in metrics['cv']]

    assert isinstance(model, Model)
    assert model.judgment_metric == metric
    assert model.cv_judgment_metric == np.mean(judge_mets)
    assert model.cv_judgment_metric_stdev == np.std(judge_mets)


def test_save_classifier(db, datarun, model, metrics):
    worker = Worker(db, datarun, models_dir=MODEL_DIR, metrics_dir=METRIC_DIR)
    hp = db.get_hyperpartitions(datarun_id=worker.datarun.id)[0]
    classifier = worker.db.start_classifier(hyperpartition_id=hp.id,
                                            datarun_id=worker.datarun.id,
                                            host='localhost',
                                            hyperparameter_values=DT_PARAMS)

    worker.db.complete_classifier = Mock()
    worker.save_classifier(classifier.id, model, metrics)
    worker.db.complete_classifier.assert_called()

    with db_session(worker.db):
        clf = db.get_classifier(classifier.id)

        loaded = load_model(clf, MODEL_DIR)
        assert isinstance(loaded, Model)
        assert loaded.method == model.method
        assert loaded.random_state == model.random_state

        assert load_metrics(clf, METRIC_DIR) == metrics


def test_is_datarun_finished(db, dataset, datarun):
    r1 = db.get_datarun(1)
    worker = Worker(db, r1)
    assert worker.is_datarun_finished()

    r2 = db.get_datarun(2)
    worker = Worker(db, r2)
    assert not worker.is_datarun_finished()

    deadline = (datetime.datetime.now() - datetime.timedelta(seconds=1)).strftime(TIME_FMT)
    worker = get_new_worker(deadline=deadline)
    assert worker.is_datarun_finished()


def test_run_classifier(worker, hyperpartition, model, metrics):
    worker.select_hyperpartition = Mock(return_value=hyperpartition)
    worker.tune_hyperparameters = Mock(return_value=DT_PARAMS)
    worker.test_classifier = Mock(return_value=(model, metrics))
    worker.save_classifier = Mock()
    worker.db = Mock()

    # make sure the function shorts out if the datarun is finished
    worker.is_datarun_finished = Mock(return_value=True)
    worker.run_classifier()
    assert not worker.select_hyperpartition.called
    assert not worker.tune_hyperparameters.called

    # make sure things run smoothly: hyperparameters are chosen and a classifier
    # is created, tested, and saved.
    worker.is_datarun_finished = Mock(return_value=False)
    worker.run_classifier()
    worker.select_hyperpartition.assert_called_once()
    worker.tune_hyperparameters.assert_called_once_with(hyperpartition)
    worker.db.start_classifier.assert_called_once_with(hyperpartition_id=hyperpartition.id,
                                                       datarun_id=worker.datarun.id,
                                                       host=ANY,
                                                       hyperparameter_values=DT_PARAMS)
    worker.test_classifier.assert_called_once_with(hyperpartition.method, DT_PARAMS)
    worker.save_classifier.assert_called_once_with(ANY, model, metrics)

    # make sure hyperpartition specification works
    hp_id = hyperpartition.id + 1
    worker.db.get_hyperpartition = lambda i: ObjWithAttrs(id=i, method='dt',
                                                          datarun_id=worker.datarun.id)
    worker.run_classifier(hyperpartition_id=hp_id)
    worker.tune_hyperparameters.assert_called_with(ObjWithAttrs(id=hp_id))

    # make sure error handling works correctly
    worker.test_classifier.side_effect = ValueError('qwerty')
    with pytest.raises(ClassifierError):
        worker.run_classifier()
    worker.db.mark_classifier_errored.assert_called_with(
        ANY, error_message=StringWith('qwerty'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

install_requires = [
    'baytune==0.2.5',
    'boto3>=1.9.146',
    'future>=0.16.0',
    'joblib>=0.11',
    'pymysql>=0.9.3',
    'cryptography>=2.6.1',
    'numpy>=1.13.1',
    'pandas>=0.22.0',
    'psutil>=5.6.1',
    'python-daemon>=2.2.3',
    'pyyaml>=3.12',
    'requests>=2.18.4',
    'scikit-learn>=0.18.2',
    'scipy>=0.19.1',
    'sklearn-pandas>=1.5.0',
    'sqlalchemy>=1.1.14',
    'flask>=1.0.2',
    'flask-restless>=0.17.0',
    'flask-sqlalchemy>=2.3.2',
    'flask-restless-swagger-2>=0.0.3',
    'simplejson>=3.16.0',
    'tqdm>=4.31.1',
]

setup_requires = [
    'pytest-runner'
]

tests_require = [
    'mock>=2.0.0',
    'pytest-cov>=2.5.1',
    'pytest-runner>=3.0',
    'pytest-xdist>=1.20.1',
    'pytest>=3.2.3',
    'google-compute-engine==2.8.12',    # required by travis
]

development_requires = [
    # general
    'bumpversion>=0.5.3',
    'pip>=9.0.1',
    'watchdog>=0.8.3',

    # docs
    'm2r>=0.2.0',
    'Sphinx>=1.7.1',
    'sphinx_rtd_theme>=0.2.4',
    'autodocsumm>=0.1.10',

    # style check
    'flake8>=3.7.7',
    'isort>=4.3.4',

    # fix style issues
    'autoflake>=1.1',
    'autopep8>=1.4.3',

    # distribute on PyPI
    'twine>=1.10.0',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1',
    'tox>=2.9.1',
]

setup(
    author="MIT Data To AI Lab",
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    description="Auto Tune Models",
    entry_points={
        'console_scripts': [
            'atm=atm.cli:main'
        ]
    },
    extras_require={
        'dev': development_requires + tests_require,
        'tests': tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    keywords='machine learning hyperparameters tuning classification',
    name='atm',
    packages=find_packages(include=['atm', 'atm.*']),
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/HDI-project/ATM',
    version='0.2.0',
    zip_safe=False,
)



from btb.tuning.tuner import BaseTuner


class MyTuner(BaseTuner):
    """
    Very bare_bones tuner that returns a random set of parameters each time.
    """

    def propose(self):
        """
        Generate and return a random set of parameters.
        """
        return self.create_candidates(1)[0, :]



from btb.selection.selector import Selector

from atm import utilities
from atm.constants import SELECTORS


def test_make_selector():
    kwargs = {
        'choices': [1, 2, 3],
        'k': 3,
        'by_algorithm': {'svm': [1, 2], 'rf': [3, 4]}
    }

    for selector_class in SELECTORS.values():
        selector = utilities.get_instance(selector_class, **kwargs)
        assert isinstance(selector, Selector)



# -*- coding: utf-8 -*-

import argparse
import logging
import multiprocessing
import os
import time

import psutil
from daemon import DaemonContext
from lockfile.pidlockfile import PIDLockFile

from atm.api import create_app
from atm.config import AWSConfig, DatasetConfig, LogConfig, RunConfig, SQLConfig
from atm.core import ATM
from atm.data import copy_files, get_demos

LOGGER = logging.getLogger(__name__)


def _get_atm(args):
    sql_conf = SQLConfig(args)
    aws_conf = AWSConfig(args)
    log_conf = LogConfig(args)

    return ATM(**sql_conf.to_dict(), **aws_conf.to_dict(), **log_conf.to_dict())


def _work(args, wait=False):
    """Creates a single worker."""
    atm = _get_atm(args)

    atm.work(
        datarun_ids=getattr(args, 'dataruns', None),
        choose_randomly=False,
        save_files=args.save_files,
        cloud_mode=args.cloud_mode,
        total_time=getattr(args, 'total_time', None),
        wait=wait
    )


def _serve(args):
    """Launch the ATM API with the given host / port."""
    atm = _get_atm(args)
    app = create_app(atm, getattr(args, 'debug', False))
    app.run(host=args.host, port=args.port)


def _get_pid_path(pid):
    """Returns abspath of the pid file which is stored on the cwd."""
    pid_path = pid

    if not os.path.isabs(pid_path):
        pid_path = os.path.join(os.getcwd(), pid_path)

    return pid_path


def _get_atm_process(pid_path):
    """Return `psutil.Process` of the `pid` file. If the pidfile is stale it will release it."""
    pid_file = PIDLockFile(pid_path, timeout=1.0)

    if pid_file.is_locked():
        pid = pid_file.read_pid()

        try:
            process = psutil.Process(pid)
            if process.name() == 'atm':
                return process
            else:
                pid_file.break_lock()

        except psutil.NoSuchProcess:
            pid_file.break_lock()


def _status(args):
    """Check if the current ATM process is runing."""

    pid_path = _get_pid_path(args.pid)
    process = _get_atm_process(pid_path)

    if process:
        workers = 0
        addr = None
        for child in process.children():
            connections = child.connections()
            if connections:
                connection = connections[0]
                addr = connection.laddr

            else:
                workers += 1

        s = 's' if workers > 1 else ''
        print('ATM is running with {} worker{}'.format(workers, s))

        if addr:
            print('ATM REST server is listening on http://{}:{}'.format(addr.ip, addr.port))

    else:
        print('ATM is not runing.')


def _start_background(args):
    """Launches the server/worker in daemon processes."""
    if args.server:
        LOGGER.info('Starting the REST API server')

        process = multiprocessing.Process(target=_serve, args=(args, ))
        process.daemon = True

        process.start()

    pool = multiprocessing.Pool(args.workers)
    for _ in range(args.workers):
        LOGGER.info('Starting background worker')
        pool.apply_async(_work, args=(args, True))

    pool.close()
    pool.join()


def _start(args):
    """Create a new process of ATM pointing the process to a certain `pid` file."""
    pid_path = _get_pid_path(args.pid)
    process = _get_atm_process(pid_path)

    if process:
        print('ATM is already running!')

    else:
        print('Starting ATM')

        if args.foreground:
            _start_background(args)

        else:
            pidfile = PIDLockFile(pid_path, timeout=1.0)

            with DaemonContext(pidfile=pidfile, working_directory=os.getcwd()):
                # Set up default log file if not already set
                if not args.logfile:
                    _logging_setup(args.verbose, 'atm.log')

                _start_background(args)


def _stop(args):
    """Stop the current running process of ATM."""
    pid_path = _get_pid_path(args.pid)
    process = _get_atm_process(pid_path)

    if process:
        process.terminate()

        for _ in range(args.timeout):
            if process.is_running():
                time.sleep(1)
            else:
                break

        if process.is_running():
            print('ATM was not able to stop after {} seconds.'.format(args.timeout))
            if args.force:
                print('Killing it.')
                process.kill()

            else:
                print('Use --force to kill it.')

        else:
            print('ATM stopped correctly.')

    else:
        print('ATM is not running.')


def _restart(args):
    _stop(args)
    time.sleep(1)

    pid_path = _get_pid_path(args.pid)
    process = _get_atm_process(pid_path)

    if process:
        print('ATM did not stop correctly. Aborting')
    else:
        _start(args)


def _enter_data(args):
    atm = _get_atm(args)
    run_conf = RunConfig(args)
    dataset_conf = DatasetConfig(args)

    if run_conf.dataset_id is None:
        dataset = atm.add_dataset(**dataset_conf.to_dict())
        run_conf.dataset_id = dataset.id

    return atm.add_datarun(**run_conf.to_dict())


def _make_config(args):
    copy_files('*.yaml', ('config'))


def _get_demos(args):
    get_demos(args)


def _get_parser():
    logging_args = argparse.ArgumentParser(add_help=False)
    logging_args.add_argument('-v', '--verbose', action='count', default=0)
    logging_args.add_argument('-l', '--logfile')

    parser = argparse.ArgumentParser(description='ATM Command Line Interface',
                                     parents=[logging_args])

    subparsers = parser.add_subparsers(title='action', help='Action to perform')
    parser.set_defaults(action=None)

    # Common Arguments
    sql_args = SQLConfig.get_parser()
    aws_args = AWSConfig.get_parser()
    log_args = LogConfig.get_parser()
    run_args = RunConfig.get_parser()
    dataset_args = DatasetConfig.get_parser()

    # Enter Data Parser
    enter_data_parents = [
        logging_args,
        sql_args,
        aws_args,
        dataset_args,
        log_args,
        run_args
    ]
    enter_data = subparsers.add_parser('enter_data', parents=enter_data_parents,
                                       help='Add a Dataset and trigger a Datarun on it.')
    enter_data.set_defaults(action=_enter_data)

    # Wroker Args
    worker_args = argparse.ArgumentParser(add_help=False)
    worker_args.add_argument('--cloud-mode', action='store_true', default=False,
                             help='Whether to run this worker in cloud mode')
    worker_args.add_argument('--no-save', dest='save_files', action='store_false',
                             help="don't save models and metrics at all")

    # Worker
    worker_parents = [
        logging_args,
        worker_args,
        sql_args,
        aws_args,
        log_args
    ]
    worker = subparsers.add_parser('worker', parents=worker_parents,
                                   help='Start a single worker in foreground.')
    worker.set_defaults(action=_work)
    worker.add_argument('--dataruns', help='Only train on dataruns with these ids', nargs='+')
    worker.add_argument('--total-time', help='Number of seconds to run worker', type=int)

    # Server Args
    server_args = argparse.ArgumentParser(add_help=False)
    server_args.add_argument('--host', help='IP to listen at')
    server_args.add_argument('--port', help='Port to listen at', type=int)

    # Server
    server = subparsers.add_parser('server', parents=[logging_args, server_args, sql_args],
                                   help='Start the REST API Server in foreground.')
    server.set_defaults(action=_serve)
    server.add_argument('--debug', help='Start in debug mode', action='store_true')
    # add_arguments_sql(server)

    # Background Args
    background_args = argparse.ArgumentParser(add_help=False)
    background_args.add_argument('--pid', help='PID file to use.', default='atm.pid')

    # Start Args
    start_args = argparse.ArgumentParser(add_help=False)
    start_args.add_argument('--foreground', action='store_true', help='Run on foreground')
    start_args.add_argument('-w', '--workers', default=1, type=int, help='Number of workers')
    start_args.add_argument('--no-server', dest='server', action='store_false',
                            help='Do not start the REST server')

    # Start
    start_parents = [
        logging_args,
        worker_args,
        server_args,
        background_args,
        start_args,
        sql_args,
        aws_args,
        log_args
    ]
    start = subparsers.add_parser('start', parents=start_parents,
                                  help='Start an ATM Local Cluster.')
    start.set_defaults(action=_start)

    # Status
    status = subparsers.add_parser('status', parents=[logging_args, background_args])
    status.set_defaults(action=_status)

    # Stop Args
    stop_args = argparse.ArgumentParser(add_help=False)
    stop_args.add_argument('-t', '--timeout', default=5, type=int,
                           help='Seconds to wait before killing the process.')
    stop_args.add_argument('-f', '--force', action='store_true',
                           help='Kill the process if it does not terminate gracefully.')

    # Stop
    stop = subparsers.add_parser('stop', parents=[logging_args, stop_args, background_args],
                                 help='Stop an ATM Local Cluster.')
    stop.set_defaults(action=_stop)

    # restart
    restart = subparsers.add_parser('restart', parents=start_parents + [stop_args],
                                    help='Restart an ATM Local Cluster.')
    restart.set_defaults(action=_restart)

    # Make Config
    make_config = subparsers.add_parser('make_config', parents=[logging_args],
                                        help='Generate a config templates folder in the cwd.')
    make_config.set_defaults(action=_make_config)

    # Get Demos
    get_demos = subparsers.add_parser('get_demos', parents=[logging_args],
                                      help='Create a demos folder and put the demo CSVs inside.')
    get_demos.set_defaults(action=_get_demos)

    return parser


def _logging_setup(verbosity=1, logfile=None):
    logger = logging.getLogger()
    log_level = (2 - verbosity) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(module)s - %(message)s'
    formatter = logging.Formatter(fmt)
    logger.setLevel(log_level)
    logger.propagate = False

    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    else:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)


def main():
    parser = _get_parser()
    args = parser.parse_args()

    _logging_setup(args.verbose, args.logfile)

    if not args.action:
        parser.print_help()
        parser.exit()

    args.action(args)



from __future__ import absolute_import, division, unicode_literals

from builtins import range

import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score, average_precision_score, cohen_kappa_score, f1_score, matthews_corrcoef,
    precision_recall_curve, roc_auc_score, roc_curve)
from sklearn.model_selection import StratifiedKFold

from atm.constants import METRICS_BINARY, METRICS_MULTICLASS, N_FOLDS_DEFAULT, Metrics


def rank_n_accuracy(y_true, y_prob_mat, n=0.33):
    """
    Compute how often the true label is one of the top n predicted classes
    for each training example.
    If n is an integer, consider the top n predictions for each example.
    If n is a float, it represents a proportion of the top predictions.
    This metric is only really useful when the total number of classes is large.
    """
    n_classes = y_prob_mat.shape[1]
    if n < 1:
        # round to nearest int before casting
        n = int(round(n_classes * n))

    # sort the rankings in descending order, then take the top n
    rankings = np.argsort(-y_prob_mat)
    rankings = rankings[:, :n]

    num_samples = len(y_true)
    correct_sample_count = 0.0   # force floating point math

    for i in range(num_samples):
        if y_true[i] in rankings[i, :]:
            correct_sample_count += 1

    return int(correct_sample_count / num_samples)


def get_per_class_matrix(y, classes=None):
    """
    Create a (num_classes x num_examples) binary matrix representation of the
    true and predicted y values.
    If classes is None, class values will be extracted from y. Values that are
    not present at all will not receive a column -- this is to allow computation
    of per-class roc_auc scores without error.
    """
    classes = classes or np.unique(y)
    y_bin = np.zeros((len(y), len(classes)))
    for i, cls in enumerate(classes):
        y_bin[:, i] = (y == cls).astype(int)
    return y_bin


def get_pr_roc_curves(y_true, y_pred_probs):
    """
    Compute precision/recall and receiver operating characteristic metrics for a
    binary class label.

    y_true: series of true class labels (only 1 or 0)
    y_pred_probs: series of probabilities generated by the model for the label
        class 1
    """
    results = {}
    roc = roc_curve(y_true, y_pred_probs, pos_label=1)
    results[Metrics.ROC_CURVE] = {
        'fprs': list(roc[0]),
        'tprs': list(roc[1]),
        'thresholds': list(roc[2]),
    }

    pr = precision_recall_curve(y_true, y_pred_probs, pos_label=1)
    results[Metrics.PR_CURVE] = {
        'precisions': list(pr[0]),
        'recalls': list(pr[1]),
        'thresholds': list(pr[2]),
    }

    return results


def get_metrics_binary(y_true, y_pred, y_pred_probs, include_curves=False):
    results = {
        Metrics.ACCURACY: accuracy_score(y_true, y_pred),
        Metrics.COHEN_KAPPA: cohen_kappa_score(y_true, y_pred),
        Metrics.F1: f1_score(y_true, y_pred),
        Metrics.MCC: matthews_corrcoef(y_true, y_pred),
        Metrics.ROC_AUC: np.nan,
        Metrics.AP: np.nan,
    }

    # if possible, compute PR and ROC curve metrics
    all_labels_same = len(np.unique(y_true)) == 1
    any_probs_nan = np.any(np.isnan(y_pred_probs))
    if not any_probs_nan:
        # AP can be computed even if all labels are the same
        y_true_bin = get_per_class_matrix(y_true, list(range(2)))
        results[Metrics.AP] = average_precision_score(y_true_bin, y_pred_probs)

        if not all_labels_same:
            results[Metrics.ROC_AUC] = roc_auc_score(y_true_bin, y_pred_probs)

        # if necessary, compute point-by-point precision/recall and ROC curve data
        if include_curves:
            results.update(get_pr_roc_curves(y_true, y_pred_probs[:, 1]))

    return results


def get_metrics_multiclass(y_true, y_pred, y_pred_probs,
                           include_per_class=False, include_curves=False):
    results = {
        Metrics.ACCURACY: accuracy_score(y_true, y_pred),
        Metrics.COHEN_KAPPA: cohen_kappa_score(y_true, y_pred),
        Metrics.F1_MICRO: f1_score(y_true, y_pred, average='micro'),
        Metrics.F1_MACRO: f1_score(y_true, y_pred, average='macro'),
        Metrics.ROC_AUC_MICRO: np.nan,
        Metrics.ROC_AUC_MACRO: np.nan,
        Metrics.RANK_ACCURACY: np.nan,
    }

    # this parameter is most relevant for datasets with high-cardinality
    # labels (lots of poosible values)
    # TODO: make the rank parameter configurable
    results[Metrics.RANK_ACCURACY] = rank_n_accuracy(y_true=y_true,
                                                     y_prob_mat=y_pred_probs)

    # if possible, compute multi-label AUC metrics
    present_classes = np.unique(y_true)
    all_labels_same = len(present_classes) == 1
    any_probs_nan = np.any(np.isnan(y_pred_probs))
    if not (all_labels_same or any_probs_nan):
        # get binary label matrix, ignoring classes that aren't present
        y_true_bin = get_per_class_matrix(y_true)

        # filter out probabilities for classes that aren't in this sample
        filtered_probs = y_pred_probs[:, present_classes]

        # actually compute roc_auc score
        results[Metrics.ROC_AUC_MICRO] = roc_auc_score(y_true_bin,
                                                       filtered_probs,
                                                       average='micro')
        results[Metrics.ROC_AUC_MACRO] = roc_auc_score(y_true_bin,
                                                       filtered_probs,
                                                       average='macro')

    # TODO: multi-label AP metrics?

    # labelwise controls whether to compute separate metrics for each posisble label
    if include_per_class or include_curves:
        results['class_wise'] = {}

        # create binary matrices, including classes that aren't actually present
        all_classes = list(range(y_pred_probs.shape[1]))
        y_true_bin = get_per_class_matrix(y_true, classes=all_classes)
        y_pred_bin = get_per_class_matrix(y_pred, classes=all_classes)

        # for each possible class, generate F1, precision-recall, and ROC scores
        # using the binary metrics function.
        for cls in all_classes:
            class_pred_probs = np.column_stack((1 - y_pred_probs[:, cls],
                                                y_pred_probs[:, cls]))
            class_res = get_metrics_binary(y_true=y_true_bin[:, cls],
                                           y_pred=y_pred_bin[:, cls],
                                           y_pred_probs=class_pred_probs,
                                           include_curves=include_curves)
            results['class_wise'][cls] = class_res

    return results


def test_pipeline(pipeline, X, y, binary, **kwargs):
    if binary:
        get_metrics = get_metrics_binary
    else:
        get_metrics = get_metrics_multiclass

    # run the test data through the trained pipeline
    y_pred = pipeline.predict(X)

    # if necessary (i.e. if a pipeline does not produce probability scores by
    # default), use class distance scores in lieu of probability scores
    method = pipeline.steps[-1][0]
    if method in ['sgd', 'pa']:
        if binary:
            class_1_distance = pipeline.decision_function(X)
            class_0_distance = -class_1_distance
            y_pred_probs = np.column_stack((class_0_distance, class_1_distance))
        else:
            y_pred_probs = pipeline.decision_function(X)
    else:
        y_pred_probs = pipeline.predict_proba(X)

    return get_metrics(y, y_pred, y_pred_probs, **kwargs)


def cross_validate_pipeline(pipeline, X, y, binary=True,
                            n_folds=N_FOLDS_DEFAULT, **kwargs):
    """
    Compute metrics for each of `n_folds` folds of the training data in (X, y).

    pipeline: the sklearn Pipeline to train and test
    X: feature matrix
    y: series of labels corresponding to rows in X
    binary: whether the label is binary or multi-ary
    n_folds: number of non-overlapping "folds" of the data to make for
        cross-validation
    """
    if binary:
        metrics = METRICS_BINARY
    else:
        metrics = METRICS_MULTICLASS

    df = pd.DataFrame(columns=metrics)
    results = []

    # TODO: how to handle classes that are so uncommon that stratified sampling
    # doesn't work? i.e. len([c for c in y if c == some_class]) < n_folds
    skf = StratifiedKFold(n_splits=n_folds)
    skf.get_n_splits(X, y)

    for train_index, test_index in skf.split(X, y):
        pipeline.fit(X[train_index], y[train_index])
        split_results = test_pipeline(pipeline=pipeline,
                                      X=X[test_index],
                                      y=y[test_index],
                                      binary=binary, **kwargs)
        df = df.append([{m: split_results.get(m) for m in metrics}])
        results.append(split_results)

    return df, results
